@@ -1,19 +1,23 @@
 # like common, but can import models
-from otree.models_concrete import PageTimeBatch
 import time
+from dataclasses import dataclass, asdict
 
-TIME_SPENT_COLUMNS = [
-    'session_code',
-    'participant_id_in_session',
-    'participant_code',
-    'page_index',
-    'app_name',
-    'page_name',
-    'epoch_time',
-    'round_number',
-    'timeout_happened',
-    'is_wait_page',
-]
+from otree.database import db
+from otree.models_concrete import PageTimeBatch
+
+
+@dataclass
+class TimeSpentRow:
+    session_code: str
+    participant_id_in_session: int
+    participant_code: str
+    page_index: int
+    app_name: str
+    page_name: str
+    epoch_time_completed: int
+    round_number: int
+    timeout_happened: int
+    is_wait_page: int
 
 
 page_completion_buffer = []
@@ -22,6 +26,18 @@ page_completion_last_write = 0
 BUFFER_SIZE = 50
 
 
+def write_row_to_page_buffer(row: TimeSpentRow):
+    d = asdict(row)
+    row = ','.join(map(str, d.values())) + '\n'
+
+    page_completion_buffer.append(row)
+    if (
+        len(page_completion_buffer) > BUFFER_SIZE
+        or time.time() - page_completion_last_write > 60 * 2
+    ):
+        write_page_completion_buffer()
+
+
 def make_page_completion_row(
     *,
     view,
@@ -32,11 +48,11 @@ def make_page_completion_row(
     is_wait_page,
 ):
     now = int(time.time())
-    fields = dict(
+    row = TimeSpentRow(
         app_name=app_name,
         page_index=view._index_in_pages,
         page_name=type(view).__name__,
-        epoch_time=now,
+        epoch_time_completed=now,
         round_number=view.round_number,
         participant_id_in_session=participant__id_in_session,
         participant_code=participant__code,
@@ -44,18 +60,11 @@ def make_page_completion_row(
         timeout_happened=int(bool(getattr(view, 'timeout_happened', False))),
         is_wait_page=is_wait_page,
     )
-    row = ','.join(str(fields[col]) for col in TIME_SPENT_COLUMNS) + '\n'
-
-    page_completion_buffer.append(row)
-    if (
-        len(page_completion_buffer) > BUFFER_SIZE
-        or now - page_completion_last_write > 60 * 2
-    ):
-        write_page_completion_buffer()
+    write_row_to_page_buffer(row)
 
 
 def write_page_completion_buffer():
     global page_completion_last_write
-    PageTimeBatch.objects.create(text=''.join(page_completion_buffer))
+    db.add(PageTimeBatch(text=''.join(page_completion_buffer)))
     page_completion_last_write = time.time()
     page_completion_buffer.clear()
